import subprocess
from typing import TypedDict
from concurrent.futures import ProcessPoolExecutor, wait, Future

import toml
import typer
from rich import print

allowedtypes = str | list[str]

tomltype = dict[str, dict[str, allowedtypes]]


class ExecutionResult(TypedDict):
    code: int
    output: str
    error: str
    command: str


def readtoml(path: str) -> tomltype:
    with open(path) as file:
        return toml.loads(file.read())


def writetoml(path: str, content: tomltype) -> bool:
    try:
        with open(path, "w") as file:
            dumped = toml.dumps(content)
            file.write(dumped)
            return True
    except:
        return False


def execute_shell_command(command: str, caputre_output: bool) -> ExecutionResult:
    result = subprocess.run(command, shell=True, capture_output=caputre_output)
    return {
        "command": result.args,
        "code": result.returncode,
        "output": result.stdout.decode("utf-8") if caputre_output else "",
        "error": result.stderr.decode("utf-8") if caputre_output else "",
    }


def execute_toml(
    commands: dict[str, allowedtypes],
    target: str,
    ignore_failed: bool,
    capture_output: bool,
    parallel: bool,
) -> list[ExecutionResult]:
    target_to_execute = commands[target]
    print_bold(f"Running [blue]'{target}'[/blue] command", "yellow")
    errors: list[ExecutionResult] = []
    execute_toml_commands(
        commands, target_to_execute, errors, ignore_failed, capture_output, parallel
    )
    return errors


def execute_toml_commands(
    commands: dict[str, allowedtypes],
    current_command: str | list[str],
    errors: list,
    ignore_failed: bool,
    capture_output: bool,
    parallel: bool,
    indent: int = 1,
):
    if isinstance(current_command, str):
        if current_command in commands:
            execute_toml_commands(
                commands,
                commands[current_command],
                errors,
                ignore_failed,
                capture_output,
                parallel,
                indent + 2,
            )
        else:
            result = execute_shell_command(current_command, capture_output)
            if result["code"] == 0:
                # Add spaces in case result is multiline
                indented_output = result["output"].replace("\n", "\n" + " " * indent)
                print_bold(indented_output, "", indent)
                print_bold(":white_heavy_check_mark: Success", "green", indent)
            else:
                print_bold(":x: Error: " + result["error"], "red", indent)
                errors.append(
                    {
                        "command": current_command,
                        "error": result["error"],
                    }
                )
                return
    else:
        if parallel:
            with ProcessPoolExecutor() as executor:
                for command in current_command:
                    executor.submit(
                        execute_toml_command,
                        commands,
                        command,
                        errors,
                        ignore_failed,
                        capture_output,
                        parallel,
                        indent,
                    )

        else:
            for command in current_command:
                execute_toml_command(
                    commands,
                    command,
                    errors,
                    ignore_failed,
                    capture_output,
                    parallel,
                    indent,
                )
            if len(errors) != 0 and not ignore_failed:
                return


def execute_toml_command(
    commands: dict[str, allowedtypes],
    command: str | list[str],
    errors: list,
    ignore_failed: bool,
    capture_output: bool,
    parallel: bool,
    indent: int = 1,
):
    print_bold(
        f"Running [blue]'{command}'[/blue] command",
        "yellow",
        indent,
    )
    execute_toml_commands(
        commands,
        command,
        errors,
        ignore_failed,
        capture_output,
        parallel,
        indent + 2,
    )


def list_to_dict(commands: list[str]) -> dict[str, str | list[str]]:
    commands_dict: dict[str, str | list[str]] = {}
    for command in commands:
        if command.count("=") != 1:
            raise typer.Abort("Please enter key=value pairs correctly.")
        key, value = command.split("=")
        if len(value) > 1 and value[0] == "[" and value[-1] == "]":  # value="[a,b]"
            value = value[1:-1]  # value="a,b"
            value_list = [c for c in value.split(",")]
            commands_dict[key] = value_list
        else:
            commands_dict[key] = value
    return commands_dict


def print_bold(content: str, color: str = "", indent=0):
    print(" " * indent + f"[bold {color}] {content} [/bold {color}]")



import typer
from rich.console import Console
from rich.table import Table

from .utils import (
    execute_toml,
    list_to_dict,
    print_bold,
    readtoml,
    tomltype,
    writetoml,
)

app = typer.Typer()

default_path = "pyproject.toml"
default_group = "yasta-tasks"


@app.command()
def init(path: str = typer.Option(default_path, "-p")):
    content: tomltype = {
        default_group: {"test": "echo 'You do not have any tests.'"},
    }
    writetoml(path, content)


@app.command()
def add(commands: list[str], path: str = typer.Option(default_path, "--path", "-p")):
    commands_dict = list_to_dict(commands)
    file = readtoml(path)
    saved_commands = file[default_group]
    for k, v in commands_dict.items():
        saved_commands[k] = v  # type: ignore
    writetoml(path, file)


@app.command()
def delete(command: str, path: str = typer.Option(default_path)):
    file = readtoml(path)
    commands = file[default_group]
    del commands[command]
    writetoml(path, file)


@app.command()
def run(
    command: str,
    path: str = typer.Option(default_path, "--path", "-p"),
    capture_output: bool = typer.Option(
        False,
        "--caputre-output",
        "-c",
        help="""
        Capturing output waits for the command to finish and then prints the output, coloring the output. \n
        This wouldn't be useful if you're running a webserver and want to see the logs.
        """,
    ),
    ignore_failed: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force running all commands by ignoring failed tests",
    ),
    parallel: bool = typer.Option(False, help="Run tasks in parallel"),
):
    file = readtoml(path)
    commands = file[default_group]
    if not isinstance(commands, dict):
        return
    errors = execute_toml(commands, command, ignore_failed, capture_output, parallel)
    if len(errors) == 0:
        print_bold(":white_heavy_check_mark: No errors occured", "green")
        return
    print_bold(f":x: {len(errors)} error(s) occured!", "red")
    table = Table(show_header=True, show_lines=True)
    table.add_column("Command")
    table.add_column("Error") if capture_output else ...
    for error in errors:
        row = {"command": f"[bold red] {error['command']} [/bold red]"}
        if capture_output:
            row["error"] = f"[bold red] {error['error']} [/bold red]"
        table.add_row(*row.values())
    console = Console()
    console.print(table)


@app.command()
def show(
    path: str = typer.Option(default_path, "--path", "-p"),
):
    file = readtoml(path)
    tasks = file[default_group]
    table = Table(show_header=True, show_lines=True)
    table.add_column("Task", style="green")
    table.add_column("Command", style="yellow")
    for k, v in tasks.items():
        table.add_row(k, v if isinstance(v, str) else str(v))
    console = Console()
    console.print(table)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yasta']

package_data = \
{'': ['*']}

install_requires = \
['toml>=0.10.2,<0.11.0', 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['yasta = yasta.yasta:app']}

setup_kwargs = {
    'name': 'yasta',
    'version': '0.1.2',
    'description': 'A modern task runner.',
    'long_description': "# Yasta\n\n![Yasta logo](https://i.imgur.com/MGw6TNI_d.webp?maxwidth=760)\n\nYasta is a modern task runner written in Python, it's also what we call microbus drivers in Egypt! 🚐\n\nYasta makes running and managing your tasks a breeze! 🌬️\n\n# How to install\n\n```\npip install yasta\n```\n\n# How to use\n\nYasta consists of 5 commands\n\n1. init (initializes a pyproject.toml file with a test command)\n2. add (adds a task to the list of tasks)\n3. delete (deletes a task from the list of tasks)\n4. show (shows the tables of tasks)\n5. run (runs a task)\n\nYou can know more about the commands and flags by running\n\n```\nyasta --help\n```\n\n# Example\n\n![Yasta init, adding and running tasks](https://i.imgur.com/Y55f9AN_d.webp?maxwidth=1520)\n\n![Yasta force running commands, ignoring failed tasks](https://i.imgur.com/FsOrx3X_d.webp?maxwidth=1520)\n",
    'author': 'Adham Salama',
    'author_email': 'adhamsalama@zohomail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/adhamsalama/yasta',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
